/**
 ** Simple entropy harvester based upon the havege RNG
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/time.h>

#include <asm/types.h>
#include <linux/random.h>
#include <errno.h>

#include "haveged.h"
/**
 * Parameters
 */
static struct pparams defaults = {
  .daemon         = "haveged",
  .is_daemon      = 1,
  .low_water      = 1024,
  .pid_file       = "/var/run/haveged.pid",
  .poolsize       = "/proc/sys/kernel/random/poolsize",
  .random_device  = "/dev/random",
  .sample_out     = "",
  .sample_size    = 1024,
  .version        = ".1",
  .watermark      = "/proc/sys/kernel/random/write_wakeup_threshold"
  };
struct pparams *params = &defaults;
/**
 * Prototypes
 */
void daemonize();
void error_exit(char *format, ...);
int  get_poosize();
void run(int poolsize,struct rand_pool_info *output, int *buffer);
void set_watermark(int level);
void tidy_exit(int signum);
/**
 * The usual daemon setup
 */
void daemonize()
{
   FILE *fh;
   openlog(params->daemon, LOG_CONS, LOG_DAEMON);
   syslog(LOG_NOTICE, "%s starting up", params->daemon);
   signal(SIGHUP, tidy_exit);
   signal(SIGINT, tidy_exit);
   signal(SIGTERM, tidy_exit);
   if (daemon(0, 0) == -1)
      error_exit("cannot fork into the background");
   fh = fopen(params->pid_file, "w");
   if (!fh)
      error_exit("Couldn't open PID file \"%s\" for writing: %m.", params->pid_file);
   fprintf(fh, "%i", getpid());
   fclose(fh);
}
/**
 * Bail....
 */
void error_exit(char *format, ...)
{
   char buffer[4096];
   va_list ap;
   va_start(ap, format);
   vsnprintf(buffer, sizeof(buffer), format, ap);
   va_end(ap);
   if (params->is_daemon>0) {
      unlink(params->pid_file);
      syslog(LOG_INFO, "%s %s", params->daemon, buffer);
      }
   else fprintf(stderr, "%s %s\n", params->daemon, buffer);
   exit(1);
}
/**
 * Get configured poolsize
 */
int get_poolsize()
{
   FILE *poolsize_fh;
   int max_bits;

   if (params->is_daemon>0) {
      poolsize_fh = fopen(params->poolsize, "rb");
      if (poolsize_fh) {
         fscanf(poolsize_fh, "%d", &max_bits);
         fclose(poolsize_fh);
         }
      else max_bits = -1;
      }
   else max_bits = 1024;
   return max_bits;
}
/**
 * Entry point
 */
int main(int argc, char **argv)
{
   static struct option long_options[] =
   {
      {"sample",     1, NULL, 's' },
      {"kb",         1, NULL, 'k' },
      {"min",        1, NULL, 'm' },
      {"help",       0, NULL, 'h' },
      {NULL,         0, NULL, 0   }
   };
   int c,poolsize;

   do {
      c = getopt_long (argc, argv, "s:m:k:hv", long_options, NULL);
      switch(c) {
         case 's':
            params->sample_out = optarg;
            params->is_daemon=0;
            break;
         case 'm':
            params->low_water = atoi(optarg);
            break;
         case 'k':
            params->sample_size = atoi(optarg);
            break;
         default:
            fprintf(stderr, "Unrecognized %c\n",c);
         case 'h':
            fprintf(stderr, "Usage: %s [options]\n\n", params->daemon);
            fprintf(stderr, "Collect entropy and feed into random pool.\n");
            fprintf(stderr, "\n");
            fprintf(stderr, "Options:\n");
            fprintf(stderr, "--kb,           -k []  Set sample size.\n");
            fprintf(stderr, "--min,          -m []  Set low water mark\n");
            fprintf(stderr, "--sample,       -s []  Do not become daemon - sample output.\n");
            fprintf(stderr, "--version,      -v     Show version\n");
            fprintf(stderr, "--help,         -h     This help.\n");
            fprintf(stderr, "\n");
            exit(1);
         case 'v':
            fprintf(stderr, "%s %s\n",params->daemon, params->version);
            exit(1);
            break;
         case -1:
            break;
         }
   } while (c!=-1);
   poolsize = get_poolsize();
   if (poolsize>0) {
      int  buf[(poolsize+sizeof(int)-1)/sizeof(int)];
      char rb[sizeof(struct rand_pool_info)+poolsize+2];
      run(poolsize,(struct rand_pool_info *)rb,buf);
      }
   else error_exit("Couldn't get poolsize");
   exit(0);
}
/**
 * The meat
 */
void run(int poolsize,struct rand_pool_info *output, int *buffer)
{
   FILE *fout;
   int random_fd = -1;
   int ct=params->sample_size*1024;

   if (params->is_daemon>0) {
      daemonize();
      if (params->low_water>0)
         set_watermark(params->low_water);
      random_fd = open(params->random_device, O_RDWR);
      if (random_fd == -1)
         error_exit("Couldn't open random device: %m");
      }
   else {
      if (!(fout = fopen (params->sample_out, "wb")))
         error_exit("cannot open file <%s> for writing.\n", params->sample_out);
      fprintf(stderr, "Writing %d byte sample\n",ct);
      }
   for(;;) {
      int current,i,nbytes,r;
      if (params->is_daemon>0) {
         fd_set write_fd;
         FD_ZERO(&write_fd);
         FD_SET(random_fd, &write_fd);
         for(;;)  {
            int rc = select(random_fd+1, NULL, &write_fd, NULL, NULL);
            if (rc >= 0) break;
            if (errno != EINTR)
               error_exit("Select error: %m");
            }
         if (ioctl(random_fd, RNDGETENTCNT, &current) == -1)
            error_exit("Couldn't query entropy-level from kernel");
         nbytes = poolsize - current;
         }
      else {
         if (ct<=0) {
            fclose(fout);
            break;
            }
         nbytes = poolsize;
         ct -= nbytes;
         }
      if(nbytes<1)   continue;
      r = (nbytes+sizeof(int)-1)/sizeof(int);
      for(i=0;i<r;i++)
         buffer[i] = ndrand();
      output->buf_size = nbytes;
      output->entropy_count = nbytes * 8;
      memcpy(output->buf, buffer, output->buf_size);
      if (params->is_daemon>0) {
         if (ioctl(random_fd, RNDADDENTROPY, output) == -1)
            error_exit("RNDADDENTROPY failed!");
         }
      else {
         if (fwrite (output->buf, 1, output->buf_size, fout) == 0)
            error_exit("cannot write data in file: %m");
         }
      }
}
/**
 * Set random write threshold
 */
void set_watermark(int level)
{
   FILE *wm_fh;

   wm_fh = fopen(params->watermark, "w");
   if (wm_fh) {
      fprintf(wm_fh, "%d\n", level);
      fclose(wm_fh);
      }
   else error_exit("set_watermark failed!");
}
/**
 * Signal handler
 */
void tidy_exit(int signum)
{
   if (params->is_daemon>0) {
      unlink(params->pid_file);
      syslog(LOG_NOTICE, "%s stopping due to signal %d", params->daemon, signum);
      }
   exit(0);
}
